//----------------------------------------------------------------------------------
// aviutl GPUtB^ T|[gNX
// Copyright (c) 2006  N099
//----------------------------------------------------------------------------------
#include <windows.h>

#include <math.h>
#include <float.h>
#include <malloc.h>
#include <mmintrin.h>
#include <xmmintrin.h>
#include <new>

#include <d3d9.h>
#include <d3dx9.h>
#include <d3dx9math.h>
#include <dxerr.h>

#include "filter.h"

#include "support.h"

//---------------------------------------------------------------------
// internal function prototypes
//---------------------------------------------------------------------
static int next_pow2(int c);
static bool isSSEusable(void);
static void yc_f16(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void f16_yc(PIXEL_YC*, const void*, void*, int, int, int, int, int);
static void yc_f32(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void f32_yc(PIXEL_YC*, const void*, void*, int, int, int, int, int);
static void yc_f16_sse(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void f16_yc_sse(PIXEL_YC*, const void*, void*, int, int, int, int, int);
static void yc_f32_sse(void*, const PIXEL_YC*, void*, int, int, int, int, int);
static void f32_yc_sse(PIXEL_YC*, const void*, void*, int, int, int, int, int);

//---------------------------------------------------------------------
// struct
//---------------------------------------------------------------------
struct PPVERT
{
    float x, y, z, rhw;
    float tu, tv;       // Texcoord for post-process source
    float tu2, tv2;     // Texcoord for the original scene

    const static D3DVERTEXELEMENT9 Decl[4];
};

const D3DVERTEXELEMENT9 PPVERT::Decl[4] =
{
	{ 0, 0,  D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITIONT, 0 },
	{ 0, 16, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,  0 },
	{ 0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,  1 },
	D3DDECL_END()
};

struct TexChain { TexChain *next; IDirect3DTexture9 *tex; };

//---------------------------------------------------------------------
// public functions
//---------------------------------------------------------------------

//RXgN^
GpuFilter::GpuFilter() :
	m_pD3D(NULL), m_pd3dDevice(NULL),
	m_pTex(NULL), m_pTexSys(NULL), m_pTexTmp(NULL),
	m_pVert(NULL), m_pVertDec(NULL),
	m_pEffect(NULL), m_pPool(NULL),
	m_pTexChain(NULL), m_pFloatTmp(NULL)
{
	m_width = m_height = m_wMax = m_wTex = m_hTex = -1;

	m_texFormat = Unknown;
	m_accType = NONE;
	m_pow2 = false;

	m_sse = isSSEusable();
}

//fXgN^
GpuFilter::~GpuFilter()
{
	//\[X̊J
	_aligned_free(m_pFloatTmp);
	SAFE_DELETE_ARRAY(m_pTexChain);
	SAFE_RELEASE(m_pPool);
	SAFE_RELEASE(m_pEffect);
	SAFE_RELEASE(m_pVert);
	SAFE_RELEASE(m_pVertDec);
	SAFE_RELEASE(m_pTexSys);
	SAFE_RELEASE(m_pTexTmp);
	SAFE_RELEASE(m_pTex);

	//AddRefĂ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);
}

//IuWFNg̏
bool GpuFilter::Init(IDirect3D9 *pD3D, IDirect3DDevice9 *pd3dDevice, const TCHAR *effect)
{
	HRESULT hr;

	//|C^̃`FbN
	if(!pd3dDevice || !pD3D) return false;

	//Ŝ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);

	//JȂ߂AddRefĂ
	m_pD3D = pD3D;
	m_pD3D->AddRef();
	m_pd3dDevice = pd3dDevice;
	m_pd3dDevice->AddRef();

	//foCX̃`FbN
	if(FAILED(CheckDeviceCap())) return false;

	//GtFNg̃[h
	hr = CreateEffect(effect);
	if(FAILED(hr)){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("GtFNg̃[hɎs܂"), TRUE);
		return false;
	}

	return true;
}

//e탊\[X̊m
bool GpuFilter::AllocResources(int w, int h, int w_max, PIXEL_FORMAT fmt, ACCESS_TYPE type, bool pow2)
{
	HRESULT hr;

	//̃`FbN
	if(!m_pEffect) return false;

	//sNZTCỸZbgi4096x4096ɐj
	if(w<0 || w>4096 || h<0 || h>4096 || w_max<0 || w_max<w){
		return false;
	}else{
		m_width = w;
		m_height = h;
		m_wMax = w_max;
	}

	//eNX`TCỸZbg
	m_texFormat = fmt;
	m_accType = type;
	m_pow2 = pow2;
	if(m_pow2){ //eNX`TCY2̗ݏɂ
		m_wTex = next_pow2(m_width);
		m_hTex = next_pow2(m_height);
	}else{
		m_wTex = m_width;
		m_hTex = m_height;
	}
	
	//ꉞJĂ
	ReleaseResources();

	//_
	hr = CreateVert(PPVERT::Decl);
	if(FAILED(hr)){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("_̃ZbgɎs܂"), TRUE);
		return false;
	}

	//eNX`
	hr = CreateTempTexture(&m_pTexTmp);
	if(FAILED(hr)){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`̍쐬Ɏs܂"), TRUE);
		return false;
	}
	hr = CreateTexture(&m_pTex, &m_pTexSys);
	if(FAILED(hr)){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`̍쐬Ɏs܂"), TRUE);
		return false;
	}

	//eNX`̃Zbg
	D3DXHANDLE h_texTgt, h_texTmp;
	h_texTgt = m_pEffect->GetParameterByName(NULL, "txTgt");
	h_texTmp = m_pEffect->GetParameterByName(NULL, "txTmp");
	if(h_texTgt){
		hr = m_pEffect->SetTexture(h_texTgt, m_pTex);
		if(FAILED(hr)){
			DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("eNX`̃ZbgɎs܂"), TRUE);
			return false;
		}
	}
	if(h_texTmp){
		hr = m_pEffect->SetTexture(h_texTmp, m_pTexTmp);
		if(FAILED(hr)){
			DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("eNX`̃ZbgɎs܂"), TRUE);
			return false;
		}
	}

	//KvȂF16pϊpe|̈쐬
	if(F16 == m_texFormat){
		m_pFloatTmp = (__m128*)_aligned_malloc(sizeof(__m128)*m_width, sizeof(__m128));
		if(NULL == m_pFloatTmp){
			DXTrace(__FILE__, (DWORD)__LINE__, E_OUTOFMEMORY, 
				TEXT("ϊp̊mۂɎs܂"), TRUE);
			return false;
		}
	}

	//eNX``F[̍쐬
	m_pTexChain = new(std::nothrow) TexChain[2];
	if(NULL == m_pTexChain){
		DXTrace(__FILE__, (DWORD)__LINE__, E_OUTOFMEMORY, TEXT("̊mۂɎs܂"), TRUE);
		return false;
	}else{
		m_pTexChain[1].tex  = m_pTex;
		m_pTexChain[1].next = &m_pTexChain[0];
		m_pTexChain[0].tex  = m_pTexTmp;
		m_pTexChain[0].next = &m_pTexChain[1];
	}

	return true;
}

//e탊\[X̊J
bool GpuFilter::ReleaseResources(void)
{	
	//\[X̊J
	_aligned_free(m_pFloatTmp);
	m_pFloatTmp = NULL;
	SAFE_DELETE_ARRAY(m_pTexChain);
	SAFE_RELEASE(m_pVertDec);
	SAFE_RELEASE(m_pTexSys);
	SAFE_RELEASE(m_pTexTmp);
	SAFE_RELEASE(m_pTex);

	return true;
}

//tB^p[^̃Zbg
bool GpuFilter::SetParams(int *param, int n)
{
	HRESULT hr;

	if(NULL==m_pEffect || NULL==param || n!= 15) return false;

	//p[^nh̎擾
	D3DXHANDLE h_ycOffs, h_ycAmp, h_bgrOffs, h_bgrAmp, h_bgrGamma;
	h_ycOffs   = m_pEffect->GetParameterByName(NULL, "ycOffs");
	h_ycAmp    = m_pEffect->GetParameterByName(NULL, "ycAmp");
	h_bgrOffs  = m_pEffect->GetParameterByName(NULL, "bgrOffs");
	h_bgrAmp   = m_pEffect->GetParameterByName(NULL, "bgrAmp");
	h_bgrGamma = m_pEffect->GetParameterByName(NULL, "bgrGamma");

	//p[^ϊ
	D3DXVECTOR4 vParams;
	//Ql
	// T_YOF, T_YGA, T_CBOF, T_CBGA, T_CROF, T_CRGA,
	// T_ROF, T_RGA, T_RGM,  T_GOF,  T_GGA,  T_GGM, T_BOF, T_BGA, T_BGM

	//yc offset
	if(h_ycOffs){
		vParams.x = 0.0f;
		vParams.y = param[0]/256.0f;
		vParams.z = param[2]/256.0f;
		vParams.w = param[4]/256.0f;
		hr = m_pEffect->SetVector(h_ycOffs, &vParams);
		if(FAILED(hr)){
			DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("p[^̃ZbgɎs܂"), TRUE);
		}
	}
	//yc gain
	if(h_ycAmp){
		vParams.x = 0.0f;
		vParams.y = (param[1]+256)/256.0f;
		vParams.z = (param[3]+256)/256.0f;
		vParams.w = (param[5]+256)/256.0f;
		hr = m_pEffect->SetVector(h_ycAmp, &vParams);
		if(FAILED(hr)){
			DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("p[^̃ZbgɎs܂"), TRUE);
		}
	}
	//rgb offset
	if(h_bgrOffs){
		vParams.x = 0.0f;
		vParams.w = param[6]/256.0f;
		vParams.z = param[9]/256.0f;
		vParams.y = param[12]/256.0f;
		hr = m_pEffect->SetVector(h_bgrOffs, &vParams);
		if(FAILED(hr)){
			DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("p[^̃ZbgɎs܂"), TRUE);
		}
	}
	//rgb gain
	if(h_bgrAmp){
		vParams.x = 0.0f;
		vParams.w = (param[7]+256)/256.0f;
		vParams.z = (param[10]+256)/256.0f;
		vParams.y = (param[13]+256)/256.0f;
		hr = m_pEffect->SetVector(h_bgrAmp, &vParams);
		if(FAILED(hr)){
			DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("p[^̃ZbgɎs܂"), TRUE);
		}
	}
	//rgb gamma
	if(h_bgrGamma){
		vParams.x = 1.0f;
		vParams.w = (float)pow(10.0, param[8]/256.0);
		vParams.z = (float)pow(10.0, param[11]/256.0);
		vParams.y = (float)pow(10.0, param[14]/256.0);
		hr = m_pEffect->SetVector(h_bgrGamma, &vParams);
		if(FAILED(hr)){
			DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("p[^̃ZbgɎs܂"), TRUE);
		}
	}

	return true;
}

//PIXEL_YCf[^foCXɓ]
bool GpuFilter::PutImage(const PIXEL_YC *data)
{
	HRESULT hr;

	if(COPY == m_accType){
		hr = PutPixelData(m_pTex, m_pTexSys, data);
	}else{
		hr = PutPixelDataDirect(m_pTex, data);
	}

	if(FAILED(hr)){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("foCXփRs[ł܂ł"), TRUE);
		return false;
	}

	return true;
}

//PIXEL_YCf[^foCX]
bool GpuFilter::GetImage(PIXEL_YC *data)
{
	HRESULT hr;

	if(COPY == m_accType){
		hr = GetPixelData(data, m_pTex, m_pTexSys);
	}else{
		hr = GetPixelDataDirect(data, m_pTex);
	}

	if(FAILED(hr)){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, 
			TEXT("foCXRs[ł܂ł"), TRUE);
		return false;
	}

	return true;
}

//f[^̏s
bool GpuFilter::ProcImage(void)
{
	if(!m_pTexChain) return false;

	HRESULT hr;
	TexChain *chain = m_pTexChain;

	//_OJn
	if(SUCCEEDED(m_pd3dDevice->BeginScene()))
	{
		m_pEffect->SetTechnique("ColorCorrection");

		// Set the vertex declaration
		m_pd3dDevice->SetVertexDeclaration(m_pVertDec);

		// Draw the quad
		UINT cPasses, p;
		m_pEffect->Begin( &cPasses, 0 );
		for(p = 0; p < cPasses; ++p){
			//RenderTarget̐ݒ
			IDirect3DTexture9 *pTarget = chain->tex;
			IDirect3DSurface9 *pTexSurf;
			hr = pTarget->GetSurfaceLevel(0, &pTexSurf);
			if(FAILED(hr)){
				DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("GetSurfaceLevel"), TRUE);
				return false;
			}
			m_pd3dDevice->SetRenderTarget(0, pTexSurf);
			pTexSurf->Release();

			//Render
			m_pEffect->BeginPass(p);
			m_pd3dDevice->SetStreamSource(0, m_pVert, 0, sizeof(PPVERT));
			m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
			m_pEffect->EndPass();

			//RT̍XV
			chain = chain->next;
		}
		m_pEffect->End();

		//End scene
		m_pd3dDevice->EndScene();
	}

	return true;
}

//---------------------------------------------------------------------
// private functions
//---------------------------------------------------------------------

//foCX̃T|[g󋵂̃`FbN
HRESULT GpuFilter::CheckDeviceCap(void)
{
	HRESULT hr;

	//|C^̃`FbN
	if(!m_pd3dDevice || !m_pD3D) return D3DERR_INVALIDCALL;

	//D3DCaps̃`FbN
	D3DCAPS9 caps;
	hr = m_pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps);
	if(D3D_OK != hr) return hr;

	//eNX`֘ÃeXg
	if(!(caps.Caps2 & D3DCAPS2_DYNAMICTEXTURES)){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("_Ci~bNeNX`pł܂"), TRUE);
		return D3DERR_NOTAVAILABLE;
	}
	if(caps.TextureCaps & D3DPTEXTURECAPS_SQUAREONLY){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("񐳕eNX`𗘗pł܂"), TRUE);
		return D3DERR_NOTAVAILABLE;
	}
	if(!(caps.TextureCaps & D3DPTEXTURECAPS_POW2)){
		; //2̗ݏȊÕTCỸeNX`
	}else if(!(caps.TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL)){
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("2̗ݏTCYȊÕeNX`͗pł܂"), TRUE);
		return D3DERR_NOTAVAILABLE;
	}

    //RTtH[}bg̃`FbN
    D3DDISPLAYMODE DisplayMode;
    hr = m_pd3dDevice->GetDisplayMode( 0, &DisplayMode );
	if(D3D_OK != hr) return hr;

    if(FAILED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_A16B16G16R16F)))
	{
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A16B16G16R16F pł܂"), TRUE);
	}else if(FAILED(m_pD3D->CheckDeviceFormat(caps.AdapterOrdinal, caps.DeviceType,
		DisplayMode.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_A32B32G32R32F)))
	{   
		DXTrace(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A32B32G32R32F pł܂"), TRUE);
	}

	return hr;
}

//_̐錾
HRESULT GpuFilter::CreateVert(const D3DVERTEXELEMENT9 *decl)
{
    //foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pVertDec);

    HRESULT hr = m_pd3dDevice->CreateVertexDeclaration(decl, &m_pVertDec);
	if(hr != D3D_OK) return hr;

	//Ŝ
	SAFE_RELEASE(m_pVert);

	const PPVERT Quad[4] =
	{
		{        -0.5f,        -0.5f, 1.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f },
		{ m_wTex -0.5f,        -0.5f, 1.0f, 1.0f, 1.0f, 0.0f, 1.0f, 0.0f },
		{        -0.5f, m_hTex -0.5f, 1.0f, 1.0f, 0.0f, 1.0f, 0.0f, 1.0f },
		{ m_wTex -0.5f, m_hTex -0.5f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f }
	};

	hr = m_pd3dDevice->CreateVertexBuffer(sizeof(PPVERT) * 4,
		D3DUSAGE_WRITEONLY|D3DUSAGE_DYNAMIC, 0, D3DPOOL_DEFAULT, &m_pVert, NULL );

	//obt@ɓ]
	LPVOID pVBData;
	if(SUCCEEDED(m_pVert->Lock(0, 0, &pVBData, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK))){
		CopyMemory(pVBData, Quad, sizeof(Quad));
		m_pVert->Unlock();
	}

	return hr;
}

//GtFNgt@CǂݍŃGtFNg쐬
HRESULT GpuFilter::CreateEffect(const TCHAR *path)
{
	const DWORD ShaderFlags = 0;
	ID3DXBuffer *pCompErr = NULL; //RpC̃G[
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pEffect);

	//GtFNgt@C̓ǂݎ
	hr = D3DXCreateEffectFromFile(m_pd3dDevice, 
		   path, NULL, NULL, ShaderFlags, m_pPool, &m_pEffect, &pCompErr);
	if(D3D_OK != hr){
		if(pCompErr){
			//RpCG[\
			char *msg = (char*)pCompErr->GetBufferPointer();
#ifdef UNICODE
			if(msg){
				size_t count = pCompErr->GetBufferSize();
				WCHAR *str = new(std::nothrow) WCHAR[count];
				if(str){
					mbstowcs(str, msg, count);
					str[count -1] = L'\0';
					DXTrace(__FILE__, (DWORD)__LINE__, hr, str, TRUE);
					delete[] str;
				}
			}
#else
			if(msg){
				DXTrace(__FILE__, (DWORD)__LINE__, hr, msg, TRUE);
			}
#endif
			pCompErr->Release();
		}
		
	}

	SAFE_RELEASE(pCompErr);
	return hr;
}

//f[^̂肪\ȃeNX`쐬
HRESULT GpuFilter::CreateTexture(IDirect3DTexture9 **ppTex, IDirect3DTexture9 **ppTexSys)
{
	HRESULT hr;
	DWORD usage = D3DUSAGE_RENDERTARGET | ((DIRECT == m_accType) ? D3DUSAGE_DYNAMIC : 0);
	D3DFORMAT fmt = (F32 == m_texFormat) ? D3DFMT_A32B32G32R32F : D3DFMT_A16B16G16R16F;

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;
	
	//Ŝ
	SAFE_RELEASE(*ppTex);

	//f[^̂̑ΏۂƂȂeNX`쐬
	hr = m_pd3dDevice->CreateTexture(m_wTex, m_hTex, 1, usage, fmt, D3DPOOL_DEFAULT, ppTex, NULL);
	if(hr != D3D_OK) return hr;

	//ԐڃANZX̂߂̃eNX`쐬
	if(DIRECT != m_accType){
		SAFE_RELEASE(*ppTexSys);
		hr = m_pd3dDevice->CreateTexture(m_wTex, m_hTex, 1, 0, fmt, D3DPOOL_SYSTEMMEM, ppTexSys, NULL);
	}

	return hr;
}

//f[^擾łȂeNX`쐬
HRESULT GpuFilter::CreateTempTexture(IDirect3DTexture9 **ppTex)
{
	DWORD usage = D3DUSAGE_RENDERTARGET;
	D3DFORMAT fmt = (F32 == m_texFormat) ? D3DFMT_A32B32G32R32F : D3DFMT_A16B16G16R16F;

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(*ppTex);

	//rfĨeNX`쐬
	return m_pd3dDevice->CreateTexture(m_wTex, m_hTex, 1, usage, fmt, D3DPOOL_DEFAULT, ppTex, NULL);
}

//_Ci~bNeNX`ւ̃f[^]
HRESULT GpuFilter::PutPixelDataDirect(IDirect3DTexture9 *dst, const PIXEL_YC *src)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	//eNX`̃bN
	hr = dst->LockRect(0, &lRect, NULL, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̓]
	if(F32 == m_texFormat){
		yc_f32(lRect.pBits, src, NULL, 
			m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_hTex);
	}else{
		yc_f16(lRect.pBits, src, m_pFloatTmp, 
			m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_hTex);
	}

	return dst->UnlockRect(0);
}

//_Ci~bNeNX`̃f[^]
// ۂ"D3DUSAGE_DYNAMIC"ȃeNX`̓_[^[Qbg
// łȂ̂ŁÅ֐͗płȂB
HRESULT GpuFilter::GetPixelDataDirect(PIXEL_YC *dst, IDirect3DTexture9 *src)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	//eNX`̃bN
	hr = src->LockRect(0, &lRect, NULL, D3DLOCK_READONLY|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̓]
	if(F32 == m_texFormat){
		f32_yc(dst, lRect.pBits, NULL, 
			m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_hTex);
	}else{
		f16_yc(dst, lRect.pBits, m_pFloatTmp, 
			m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_hTex);
	}

	return src->UnlockRect(0);
}

//UpdateTexture𗘗pԐړ]
HRESULT GpuFilter::PutPixelData(IDirect3DTexture9 *dst, IDirect3DTexture9 *tmp, const PIXEL_YC *src)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	//VXẽeNX`̃bN
	hr = tmp->LockRect(0, &lRect, NULL, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̃Rs[
	if(F32 == m_texFormat){
		if(m_sse){
			yc_f32_sse(lRect.pBits, src, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_hTex);
		}else{
			yc_f32(lRect.pBits, src, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_hTex);
		}
	}else{
		if(m_sse){
			yc_f16_sse(lRect.pBits, src, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_hTex);
		}else{
			yc_f16(lRect.pBits, src, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_hTex);
		}
	}
	tmp->UnlockRect(0);

	//f[^̓]
	return m_pd3dDevice->UpdateTexture(tmp, dst);
}

//GetRenderTargetData𗘗pԐړ]
HRESULT GpuFilter::GetPixelData(PIXEL_YC *dst, IDirect3DTexture9 *src, IDirect3DTexture9 *tmp)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;
	IDirect3DSurface9 *surfSrc = NULL;
	IDirect3DSurface9 *surfTmp = NULL;

	//T[tFX擾
	hr = src->GetSurfaceLevel(0, &surfSrc);
	if(hr != D3D_OK) return hr;

	hr = tmp->GetSurfaceLevel(0, &surfTmp);
	if(hr != D3D_OK){
		SAFE_RELEASE(surfSrc);
		return hr;
	}

	//VXeɓ]
	hr = m_pd3dDevice->GetRenderTargetData(surfSrc, surfTmp);
	SAFE_RELEASE(surfTmp);
	SAFE_RELEASE(surfSrc);
	if(hr != D3D_OK) return hr;

	//VXẽeNX`̃bN
	hr = tmp->LockRect(0, &lRect, NULL, D3DLOCK_READONLY|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̃Rs[
	if(F32 == m_texFormat){
		if(m_sse){
			f32_yc_sse(dst, lRect.pBits, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_hTex);
		}else{
			f32_yc(dst, lRect.pBits, NULL, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m128), m_hTex);
		}
	}else{
		if(m_sse){
			f16_yc_sse(dst, lRect.pBits, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_hTex);
		}else{
			f16_yc(dst, lRect.pBits, m_pFloatTmp, 
				m_width, m_height, m_wMax, lRect.Pitch/sizeof(__m64), m_hTex);
		}
	}

	return tmp->UnlockRect(0);
}

//---------------------------------------------------------------------
// internal functions
//---------------------------------------------------------------------
//PIXEL_YC̕ϊ
// Y (0, 4096) <-> (0.0, 1.0)
// C (-2048, 2048) <-> (-0.5, 0.5)

//t[̃TCY
// PIXEL_YC
//        w
//  ==============----- 
//  *            *    |
// h*            *    |
//  *            *    |max_h
//  *============*    |
//  |       max_w     |
//  -------------------
//
// Texture
//        w
//  ==============------- 
//  *            *      |
// h*            *      |
//  *            *      |
//  *============*      |sh
//  |                   |
//  |        sw         |
//  ---------------------
//
// float -> short ̊ۂ߂ /QIfist IvV𗘗pivs2005ł͈ӖHj 
// SSE𗘗p邱Ƃł؎̂Ăł(CVTPS2PI)

//ϊRs[ PIXEL_YC -> A16B16G16R16F
void yc_f16(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	const float r = 1.0f/4096.0f;
	float *ftmp = (float*)tmp;
	D3DXFLOAT16 *stmp = (D3DXFLOAT16*)dst;

	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			ftmp[j*4] = 0.0f;
			ftmp[j*4 + 1] = src->y  * r;
			ftmp[j*4 + 2] = src->cb * r;
			ftmp[j*4 + 3] = src->cr * r;
			src++;
		}
		D3DXFloat32To16Array(stmp, ftmp, w*4);
		stmp += w*4;
		//ZeroMemory(stmp, (sw-w)*sizeof(D3DXFLOAT16)*4);

		src += (max_w - w);
		stmp += (sw - w)*4;
	}
	//ZeroMemory(stmp, sw*(sh-h)*sizeof(D3DXFLOAT16)*4);
}

//ϊRs[ A16B16G16R16F -> PIXEL_YC
void f16_yc(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	float *ftmp = (float*)tmp;
	const D3DXFLOAT16 *stmp = (const D3DXFLOAT16*)src;
	
	for(i=0; i<h; ++i){
		D3DXFloat16To32Array(ftmp, stmp, w*4);
		for(j=0; j<w; ++j){
			dst->y  = (short)(4096.0f * ftmp[j*4 + 1]);
			dst->cb = (short)(4096.0f * ftmp[j*4 + 2]);
			dst->cr = (short)(4096.0f * ftmp[j*4 + 3]);
			dst++;
		}
		stmp += sw*4;
		dst += (max_w - w);
	}
}

//ϊRs[ PIXEL_YC -> A32B32G32R32F
void yc_f32(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	const float r = 1.0f/4096.0f;
	float *fdst = (float*)dst;

	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			*fdst++ = 0.0f;
			*fdst++ = src->y  * r;
			*fdst++ = src->cb * r;
			*fdst++ = src->cr * r;
			src++;
		}
		//ZeroMemory(fdst, (sw-w)*sizeof(float)*4);

		src += (max_w - w);
		fdst += (sw - w)*4;
	}
	//ZeroMemory(fdst, sw*(sh-h)*sizeof(float)*4);
}

//ϊRs[ A32B32G32R32F -> PIXEL_YC
void f32_yc(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	const float *fsrc = (const float*)src;
	
	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			dst->y  = (short)(4096.0f * fsrc[j*4 + 1]);
			dst->cb = (short)(4096.0f * fsrc[j*4 + 2]);
			dst->cr = (short)(4096.0f * fsrc[j*4 + 3]);
			dst++;
		}
		fsrc += sw*4;
		dst += (max_w - w);
	}

}

//ϊRs[ PIXEL_YC -> A16B16G16R16F SSEp
void yc_f16_sse(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	__m64 m0;
	__m128 x0;
	float *ftmp = (float*)tmp;
	static const __m128 xr = { 0.0f, 1.0f/4096.0f, 1.0f/4096.0f, 1.0f/4096.0f };
	D3DXFLOAT16 *stmp = (D3DXFLOAT16*)dst;

	for(i=0; i<h; ++i){
		m0 = *(__m64*)src;
		for(j=0; j<w; ++j){
			m0 = _mm_slli_si64(m0, 16);
			x0 = _mm_cvtpi16_ps(m0);

			src++;
			m0 = *(__m64*)src;

			x0 = _mm_mul_ps(x0, xr);
			_mm_store_ps(ftmp + 4*j, x0);
		}
		D3DXFloat32To16Array(stmp, ftmp, w*4);

		src += (max_w - w);
		stmp += sw*4;
	}

	_mm_empty();
}

//ϊRs[ A16B16G16R16F -> PIXEL_YC SSEp
void f16_yc_sse(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	__m64 m0;
	__m128 x0;
	float *ftmp = (float*)tmp;
	static const __m128 xc = { 0.0f, 4096.0f, 4096.0f, 4096.0f };
	const D3DXFLOAT16 *stmp = (const D3DXFLOAT16*)src;
	
	for(i=0; i<h; ++i){
		D3DXFloat16To32Array(ftmp, stmp, w*4);
		for(j=0; j<w; ++j){
			x0 = _mm_load_ps(ftmp + 4*j);
			x0 = _mm_mul_ps(x0, xc);

			m0 = _mm_cvtps_pi16(x0);
			m0 = _mm_srli_si64(m0, 16);
			*(__m64*)dst = m0;
			++dst;
		}
		stmp += sw*4;
		dst += (max_w - w);
	}

	_mm_empty();
}

//ϊRs[ PIXEL_YC -> A32B32G32R32F SSEp
void yc_f32_sse(void *dst, const PIXEL_YC *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
#if 1
	int i,j;
	__m64 m0;
	__m128 x0;
	__m128 *fdst = reinterpret_cast<__m128*>(dst);
	static const __m128 xr = { 0.0f, 1.0f/4096.0f, 1.0f/4096.0f, 1.0f/4096.0f };

	for(i=0; i<h; ++i){
		m0 = *(__m64*)src;
		for(j=0; j<w; ++j){
			m0 = _mm_slli_si64(m0, 16);
			x0 = _mm_cvtpi16_ps(m0);

			src++;
			m0 = *(__m64*)src;

			x0 = _mm_mul_ps(x0, xr);
			//_mm_store_ps((float*)(fdst + j), x0);
			_mm_stream_ps((float*)(fdst + j), x0);
		}
		src += (max_w - w);
		fdst += sw;
	}
#else //ǂݍݎ̃ACg킹ǂ܂ʂȂ
	int i,j;
	__m64 m0,m1,m2,m3;
	__m128 x0,x1,x2,x3;
	__m64 *msrc;
	__m128 *fdst = (__m128*)dst;
	static const __m128 xr = { 0.0f, 1.0f/4096.0f, 1.0f/4096.0f, 1.0f/4096.0f };

	for(i=0; i<h; ++i){
		msrc = (__m64*)src;
		m0 = *msrc++;
		for(j=0; j<w; j+=4){
			m1 = *msrc++;

			m3 = _mm_slli_si64(m0, 16);
			x0 = _mm_cvtpi16_ps(m3);

			x0 = _mm_mul_ps(x0, xr);
			_mm_stream_ps((float*)(fdst + j), x0);

			m0 = _mm_srli_si64(m0, 32);
			m0 = _mm_unpacklo_pi32(m0, m1);
			x1 = _mm_cvtpi16_ps(m0);
			m2 = *msrc++;

			x1 = _mm_mul_ps(x1, xr);
			_mm_stream_ps((float*)(fdst + j + 1), x1);

			m1 = _mm_srli_si64(m1, 32);
			m1 = _mm_unpacklo_pi32(m1, m2);
			m1 = _mm_slli_si64(m1, 16);
			x2 = _mm_cvtpi16_ps(m1);
			m0 = *msrc++;
			
			x2 = _mm_mul_ps(x2, xr);
			_mm_stream_ps((float*)(fdst + j + 2), x2);
			
			x3 = _mm_cvtpi16_ps(m2);

			x3 = _mm_mul_ps(x3, xr);
			_mm_stream_ps((float*)(fdst + j + 3), x3);
		}
		src += max_w;
		fdst += sw;
	}
#endif
	_mm_empty();
}

//ϊRs[ A32B32G32R32F -> PIXEL_YC SSEp
void f32_yc_sse(PIXEL_YC *dst, const void *src, void *tmp, int w, int h, int max_w, int sw, int sh)
{
	int i,j;
	__m64 m0;
	__m128 x0;
	const __m128 *fsrc = (const __m128*)src;
	static const __m128 xc = { 0.0f, 4096.0f, 4096.0f, 4096.0f };
	
	for(i=0; i<h; ++i){
		for(j=0; j<w; ++j){
			x0 = _mm_load_ps((float*)(fsrc + j));
			x0 = _mm_mul_ps(x0, xc);

			m0 = _mm_cvtps_pi16(x0);
			m0 = _mm_srli_si64(m0, 16);
			*(__m64*)dst = m0;
			++dst;
		}
		fsrc += sw;
		dst += (max_w - w);
	}

	_mm_empty();
}

//2̗ݏłȂꍇA2̗ݏɐ؂グ(max:4096)
static int next_pow2(int c)
{
	int i = 0;

	if(0 < c){
		for(i=1; i<4096; i<<=1){
			if(i >= c) break;
		}
	}

	return i;
}

//SSEɑΉĂ邩`FbN
static bool isSSEusable(void)
{
	long r_ecx,r_edx;

	// CPUIDɂ`FbN
	__asm {
		mov eax, 1
		cpuid
		mov r_edx, edx
		mov r_ecx, ecx
	}

	// MMX  0x0800000 edx
	// SSE  0x2000000 edx
	// SSE2 0x4000000 edx
	// SSE3 0x0000001 ecx
	r_edx &= 0x6800000;
	r_ecx &= 0x0000001;

	if(r_edx < 0x1000000)
	return false;

	// OST|[gĂ邩
	__try {
		__asm xorps xmm0, xmm0
	}

	__except(EXCEPTION_EXECUTE_HANDLER){
		if (_exception_code()==STATUS_ILLEGAL_INSTRUCTION)
			/* SSE not supported */
			return false;
	}
	/* SSE are supported by OS */
	return true;
}
